/* -*-c-*-
 *
 * $Id: keymgmt.c,v 1.2 2001/06/19 22:07:09 mdw Exp $
 *
 * Key loading and storing
 *
 * (c) 2001 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Trivial IP Encryption (TrIPE).
 *
 * TrIPE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * TrIPE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with TrIPE; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: keymgmt.c,v $
 * Revision 1.2  2001/06/19 22:07:09  mdw
 * Cosmetic fixes.
 *
 * Revision 1.1  2001/02/03 20:26:37  mdw
 * Initial checkin.
 *
 */

/*----- Header files ------------------------------------------------------*/

#include "tripe.h"

/*----- Global variables --------------------------------------------------*/

mpmont mg;
dh_priv kpriv;

/*----- Static variables --------------------------------------------------*/

static key_file *kf_pub;
static const char *kr_priv, *kr_pub, *tag_priv;
static fwatch w_priv, w_pub;

/*----- Main code ---------------------------------------------------------*/

/* --- @keymoan@ --- *
 *
 * Arguments:	@const char *file@ = name of the file
 *              @int line@ = line number in file
 *              @const char *msg@ = error message
 *              @void *p@ = argument pointer
 *
 * Returns:     ---
 *
 * Use:         Reports an error message about loading a key file.
 */

static void keymoan(const char *file, int line, const char *msg, void *p)
{
  a_warn("%s:%i: error: %s", file, line, msg);
}

/* --- @loadpriv@ --- *
 *
 * Arguments:	@dstr *d@ = string to write errors in
 *		@dh_priv *dh@ = where to store the key
 *
 * Returns:	Zero if OK, nonzero on error.
 *
 * Use:		Loads the private key from its keyfile.
 */

static int loadpriv(dstr *d, dh_priv *dh)
{
  key_file kf;
  key_packstruct kps[DH_PRIVFETCHSZ];
  key_packdef *kp;
  dh_priv mydh;
  int rc = 0;
  int e;

  if (key_open(&kf, kr_priv, KOPEN_READ, keymoan, 0)) {
    dstr_putf(d, "error reading private keyring `%s': %s",
	      kr_priv, strerror(errno));
    rc = -1;
  } else {
    T( trace(T_KEYMGMT, "keymgmt: loaded private keyring `%s'", kr_priv); )
    kp = key_fetchinit(dh_privfetch, kps, &mydh);
    if ((e = key_fetchbyname(kp, &kf, tag_priv)) != 0) {
      dstr_putf(d, "error loading private key `%s': %s",
		tag_priv, key_strerror(e));
      rc = -1;
    } else {
      dh->dp.p = MP_COPY(mydh.dp.p);
      dh->dp.q = MP_COPY(mydh.dp.q);
      dh->dp.g = MP_COPY(mydh.dp.g);
      dh->x = MP_COPY(mydh.x);
      dh->y = MP_COPY(mydh.y);      
      IF_TRACING(T_KEYMGMT, {
	trace(T_KEYMGMT, "keymgmt: extracted private key `%s'", tag_priv);
	IF_TRACING(T_CRYPTO, {
	  trace(T_CRYPTO, "crypto: p = %s", mpstr(kpriv.dp.p));
	  trace(T_CRYPTO, "crypto: q = %s", mpstr(kpriv.dp.q));
	  trace(T_CRYPTO, "crypto: g = %s", mpstr(kpriv.dp.g));
	  trace(T_CRYPTO, "crypto: x = %s", mpstr(kpriv.x));
	  trace(T_CRYPTO, "crypto: g^x = %s", mpstr(kpriv.y));
	})
      })
    }
    key_fetchdone(kp);
    key_close(&kf);
  }
  return (rc);
}

/* --- @loadpub@ --- *
 *
 * Arguments:	@dstr *d@ = string to write errors to
 *
 * Returns:	Zero if OK, nonzero on error.
 *
 * Use:		Reloads the public keyring.
 */

static int loadpub(dstr *d)
{
  key_file *kf = CREATE(key_file);

  if (key_open(kf, kr_pub, KOPEN_READ, keymoan, 0)) {
    dstr_putf(d, "error reading public keyring `%s': %s",
	      kr_pub, strerror(errno));
    DESTROY(kf);
    return (-1);
  }
  kf_pub = kf;
  T( trace(T_KEYMGMT, "keymgmt: loaded public keyring `%s'", kr_pub); )
  return (0);
}

/* --- @km_interval@ --- *
 *
 * Arguments:	---
 *
 * Returns:	Zero if OK, nonzero to force reloading of keys.
 *
 * Use:		Called on the interval timer to perform various useful jobs.
 */

int km_interval(void)
{
  dstr d = DSTR_INIT;
  dh_priv dh;
  key_file *kf;
  int reload = 0;

  /* --- Check the private key first --- */

  if (fwatch_update(&w_priv, kr_priv)) {
    T( trace(T_KEYMGMT, "keymgmt: private keyring updated: reloading..."); )
    DRESET(&d);
    if (loadpriv(&d, &dh))
      a_warn("%s -- ignoring changes", d.buf);
    else {
      reload = 1;
      mpmont_destroy(&mg);
      dh_privfree(&kpriv);
      kpriv = dh;
      mpmont_create(&mg, kpriv.dp.p);
    }
  }

  /* --- Now check the public keys --- */

  if (fwatch_update(&w_pub, kr_pub)) {
    T( trace(T_KEYMGMT, "keymgmt: public keyring updated: reloading..."); )
    kf = kf_pub;
    DRESET(&d);
    if (loadpub(&d))
      a_warn("%s -- ignoring changes", d.buf);
    else {
      reload = 1;
      key_close(kf);
      DESTROY(kf);
    }
  }

  /* --- Done --- */

  return (reload);
}

/* --- @km_init@ --- *
 *
 * Arguments:	@const char *priv@ = private keyring file
 *		@const char *pub@ = public keyring file
 *		@const char *tag@ = tag to load
 *
 * Returns:	---
 *
 * Use:		Initializes, and loads the private key.
 */

void km_init(const char *priv, const char *pub, const char *tag)
{
  dstr d = DSTR_INIT;

  kr_priv = priv;
  kr_pub = pub;
  tag_priv = tag;
  fwatch_init(&w_priv, kr_priv);
  fwatch_init(&w_pub, kr_pub);

  DRESET(&d);
  if (loadpriv(&d, &kpriv))
    die(EXIT_FAILURE, "%s", d.buf);
  mpmont_create(&mg, kpriv.dp.p);
  if (loadpub(&d))
    die(EXIT_FAILURE, "%s", d.buf);
}

/* --- @km_getpubkey@ --- *
 *
 * Arguments:	@const char *tag@ = public key tag to load
 *		@dh_pub *kpub@ = where to put the public key
 *
 * Returns:	Zero if OK, nonzero if it failed.
 *
 * Use:		Fetches a public key from the keyring.
 */

int km_getpubkey(const char *tag, dh_pub *kpub)
{
  key_packstruct kps[DH_PUBFETCHSZ];
  key_packdef *kp;
  dh_pub dp;
  int e;

  kp = key_fetchinit(dh_pubfetch, kps, &dp);
  e = key_fetchbyname(kp, kf_pub, tag);
  key_fetchdone(kp);
  if (e) {
    a_warn("error loading public key `%s': %s", tag, key_strerror(e));
    return (-1);
  }
  IF_TRACING(T_KEYMGMT, {
    trace(T_KEYMGMT, "keymgmt: extracted public key `%s'", tag);
    IF_TRACING(T_CRYPTO, {
      trace(T_CRYPTO, "crypto: p = %s", mpstr(dp.dp.p));
      trace(T_CRYPTO, "crypto: q = %s", mpstr(dp.dp.q));
      trace(T_CRYPTO, "crypto: g = %s", mpstr(dp.dp.g));
      trace(T_CRYPTO, "crypto: g^x = %s", mpstr(dp.y));
    })
  })
  if (!mp_eq(dp.dp.p, kpriv.dp.p) ||
      !mp_eq(dp.dp.q, kpriv.dp.q) ||
      !mp_eq(dp.dp.g, kpriv.dp.g)) {
    a_warn("public key `%s' has different group from private key", tag);
    return (-1);
  }
  kpub->dp.p = MP_COPY(dp.dp.p);
  kpub->dp.q = MP_COPY(dp.dp.q);
  kpub->dp.g = MP_COPY(dp.dp.g);
  kpub->y = MP_COPY(dp.y);
  return (0);
}

/*----- That's all, folks -------------------------------------------------*/
